/* winspector.c - window attribute inspector
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "screen.h"
#include "wcore.h"
#include "framewin.h"
#include "window.h"
#include "workspace.h"
#include "funcs.h"
#include "defaults.h"
#include "dialog.h"
#include "icon.h"
#include "stacking.h"
#include "application.h"
#include "appicon.h"
#include "winspector.h"

#include <proplist.h>

extern WPreferences wPreferences;

static InspectorPanel *panelList=NULL;


#define PWIDTH	260
#define PHEIGHT	350

#define COUNTARG(table)		(sizeof(table)/sizeof(WMArgument))


static WMArgument panelArgs[] = {
    {WARG_WIDTH,		(WMValue)PWIDTH},
    {WARG_HEIGHT,		(WMValue)PHEIGHT}
};

static WMArgument buttonArgs[] = {
    {WARG_X,			(WMValue)0},
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y,			(WMValue)310},
    {WARG_WIDTH,		(WMValue)62}
};

static WMArgument pageBtnArgs[] = {
    {WARG_X,			(WMValue)0},
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y,			(WMValue)10},
    {WARG_WIDTH,		(WMValue)30},
    {WARG_HEIGHT,		(WMValue)30},
    {WARG_TAG,			(WMValue)1}
};

static WMArgument specFrmArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_X,			(WMValue)40},
    {WARG_Y, 			(WMValue)65},
    {WARG_WIDTH,		(WMValue)180},
    {WARG_HEIGHT,		(WMValue)85}
};

static WMArgument specLblArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_ALIGNMENT,		(WMValue)WACenter},
    {WARG_X,			(WMValue)20},
    {WARG_Y,			(WMValue)170},
    {WARG_WIDTH,		(WMValue)210},
    {WARG_HEIGHT,		(WMValue)100}
};

static WMArgument specRadArgs[] = {
    {WARG_Y,			(WMValue)0},
    {WARG_CAPTION, 		(WMValue)NULL},
    {WARG_SELECTED,		(WMValue)1},
    {WARG_X, 			(WMValue)10},
    {WARG_WIDTH,		(WMValue)160},
    {WARG_TAG,			(WMValue)1}
};


static WMArgument attrFrmArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_X,			(WMValue)20},
    {WARG_Y, 			(WMValue)65},
    {WARG_WIDTH,		(WMValue)220},
    {WARG_HEIGHT,		(WMValue)210}
};

static WMArgument attrChkArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y, 			(WMValue)20},
    {WARG_SELECTED,		(WMValue)0},
    {WARG_X,			(WMValue)5},
    {WARG_WIDTH,		(WMValue)205},
    {WARG_HEIGHT,		(WMValue)20}
};


static WMArgument iconFrmArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y, 			(WMValue)50},
    {WARG_X,			(WMValue)15},
    {WARG_WIDTH,		(WMValue)230},
    {WARG_HEIGHT,		(WMValue)170}
};

static WMArgument wsFrmArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y, 			(WMValue)225},
    {WARG_X,			(WMValue)15},
    {WARG_WIDTH,		(WMValue)230},
    {WARG_HEIGHT,		(WMValue)70}
};


static WMArgument iconLblArgs[] = {
    {WARG_X,			(WMValue)130},
    {WARG_Y, 			(WMValue)30},
    {WARG_WIDTH,		(WMValue)64},
    {WARG_HEIGHT,		(WMValue)64},
    {WARG_RELIEF,		(WMValue)WRRaised},
    {WARG_IMAGE_POSITION,	(WMValue)WIPImageOnly}
};


static WMArgument updateIconBtnArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_Y, 			(WMValue)30},
    {WARG_X,			(WMValue)22},
    {WARG_WIDTH,		(WMValue)70}
};

static WMArgument fileLblArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_X,			(WMValue)20},
    {WARG_Y, 			(WMValue)95},
    {WARG_WIDTH,		(WMValue)104}
};

static WMArgument fileArgs[] = {
    {WARG_TEXT,			(WMValue)NULL},
    {WARG_X,			(WMValue)20},
    {WARG_Y, 			(WMValue)115},
    {WARG_WIDTH,		(WMValue)186}
};

static WMArgument alwChkArgs[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_SELECTED,		(WMValue)0},
    {WARG_X,			(WMValue)20},
    {WARG_Y, 			(WMValue)140},
    {WARG_WIDTH,		(WMValue)186}
};



static WMArgument workspaceArgs[] = {
    {WARG_TEXT,			(WMValue)NULL},
    {WARG_X,			(WMValue)30},
    {WARG_Y, 			(WMValue)40},
    {WARG_WIDTH,		(WMValue)170}
};

static WMArgument wsRad1Args[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_SELECTED,		(WMValue)0},
    {WARG_Y, 			(WMValue)15},
    {WARG_X,			(WMValue)10},
    {WARG_WIDTH,		(WMValue)180},
    {WARG_TAG,			(WMValue)1}
};

static WMArgument wsRad2Args[] = {
    {WARG_CAPTION,		(WMValue)NULL},
    {WARG_SELECTED,		(WMValue)0},
    {WARG_Y, 			(WMValue)40},
    {WARG_X,			(WMValue)10},
    {WARG_WIDTH,		(WMValue)25},
    {WARG_TAG,			(WMValue)1}
};



static void
destroyInspector(WCoreWindow *foo, void *data, XEvent *event)
{
    InspectorPanel *panel;
    InspectorPanel *tmp;
    
    panel = panelList;
    while (panel->frame!=data)
	panel = panel->nextPtr;

    WMDestroyWidget(panel->win);
    wUnmanageWindow(panel->frame, False);

    panel->inspected->flags.inspector_open = 0;

    if (panelList == panel) 
	panelList = panel->nextPtr;
    else {
	tmp = panelList;
	while (tmp->nextPtr!=panel) {
	    tmp = tmp->nextPtr;
	}
	tmp->nextPtr = panel->nextPtr;
    }

    panel->inspected->inspector = NULL;
    free(panel);
}



static void
changePage(WMButton *button, InspectorPanel *panel)
{
    if (button == panel->pageBtn1) {
	WMMapWidget(panel->specFrm);
	WMMapWidget(panel->specLbl);
    } else if (button == panel->pageBtn2) {
	WMMapWidget(panel->attrFrm);	
    } else if (button == panel->pageBtn3) {	
	WMMapWidget(panel->moreFrm);
    } else {
	WMMapWidget(panel->iconFrm);
	WMMapWidget(panel->wsFrm);
    }
    
    if (button!=panel->pageBtn1) {
	WMUnmapWidget(panel->specFrm);
	WMUnmapWidget(panel->specLbl);
    }
    if (button!=panel->pageBtn2)
	WMUnmapWidget(panel->attrFrm);
    if (button!=panel->pageBtn3)
	WMUnmapWidget(panel->moreFrm);
    if (button!=panel->pageBtn4) {
	WMUnmapWidget(panel->iconFrm);
	WMUnmapWidget(panel->wsFrm);
    }
}


#define USE_TEXT_FIELD       1
#define UPDATE_TEXT_FIELD    2

static void
showIconFor(WMScreen *scrPtr, InspectorPanel *panel,
            char *wm_instance, char *wm_class, int flags)
{
    WMPixmap *pixmap;
    WMArgument arg;
    char *file=NULL, *path;
    char *db_icon = wDefaultGetIconFile(panel->inspected->screen_ptr,
                                         wm_instance, wm_class);

    if ((flags & USE_TEXT_FIELD) != 0) {
        file = WMGetTextFromTextField(panel->fileText);

        /* If textfield is empty, try to use default icon if available */
        if (file == NULL && db_icon != NULL) {
            file = strdup(db_icon);
            flags |= UPDATE_TEXT_FIELD;
        }
    }
    else if(db_icon != NULL)
        file = strdup(db_icon);
    
    if (!file)
        return;

    if ((flags & UPDATE_TEXT_FIELD) != 0) {
        fileArgs[0].value = (WMValue)file;
        WMConfigureTextField(panel->fileText, fileArgs, COUNTARG(fileArgs));
    }

    path = FindImage(wPreferences.pixmap_path, file);
    if (file)
        free(file);

    if (!path) {
        wMessageDialog(panel->frame->screen_ptr, _("Error"),
                       _("Could not find specified icon file"), WD_ERROR);
        return;
    }
    pixmap = WMCreatePixmapFromFile(scrPtr, path);
    free(path);
    
    if (!pixmap) {
        wMessageDialog(panel->frame->screen_ptr, _("Error"),
                       _("Could not open specified icon file"), WD_ERROR);
        /*
         WMRunAlertPanel(scrPtr, _("Error"),
         _("Could not open icon file"), _("OK"), NULL, NULL);
         */
        return;
    }
    
    arg.argument = WARG_IMAGE;
    arg.value = (WMValue)pixmap;
    WMConfigureLabel(panel->iconLbl, &arg, 1);
    WMReleasePixmap(pixmap);
}


static void
updateIcon(WMButton *button, InspectorPanel *panel)
{
    showIconFor(W_SCREEN(button), panel, NULL, NULL, USE_TEXT_FIELD);
}


static void
defaultIcon(WMButton *button, InspectorPanel *panel)
{
    char *icon = NULL;
    WDDomain *db = panel->frame->screen_ptr->dwindow;
    proplist_t key, attr, value, icon_value;
    proplist_t dict = db->dictionary;

    showIconFor(W_SCREEN(button), panel, NULL, NULL, USE_TEXT_FIELD);

    if (!dict) {
        dict = PLMakeDictionaryFromEntries(NULL, NULL, NULL);
        db->dictionary = dict;
        if (dict) {
            proplist_t tmp = PLMakeString(db->path);
            PLSetFilename(dict, tmp);
            PLRelease(tmp);
        }
    }

    icon = WMGetTextFromTextField(panel->fileText);

    if (!icon || !dict)
        return;

    key = PLMakeString("*");
    attr = PLMakeString("Icon");
    value = PLMakeString(icon);
    icon_value = PLMakeDictionaryFromEntries(attr, value, NULL);
    PLRelease(attr);
    PLRelease(value);
    attr = PLGetDictionaryEntry(dict, key);
    if(attr) {
        if (PLIsDictionary(attr))
            PLMergeDictionaries(attr, icon_value);
    }
    else {
        PLInsertDictionaryEntry(dict, key, icon_value);
    }
    PLRelease(key);
    PLRelease(icon_value);
    PLSave(dict, YES);
    free(icon);
}


/*
 * Will insert the attribute = value; pair in window's list,
 * if it's different from the defaults.
 */


static void
insertAttribute(proplist_t def_win, proplist_t window, char *attr_name,
                proplist_t value, int always, int *modified)
{
    proplist_t attr = PLMakeString(attr_name);
    proplist_t def_attr = (proplist_t) NULL;
    int same = 0;
    char *val = NULL;
    char *def_val = NULL;

    if (PLIsString(value))
        val = PLGetString(value);
    if (def_win) {
        def_attr = PLGetDictionaryEntry(def_win, attr);
        if (def_attr && PLIsString(def_attr))
            def_val = PLGetString(def_attr);
    }
    if (val && def_val)
        same = !strcasecmp(val, def_val);

    if (always || !same) {
        PLInsertDictionaryEntry(window, attr, value);
        *modified = 1;
    }
    PLRelease(attr);
}


static void
saveSettings(WMButton *button, InspectorPanel *panel)
{
    WWindow *wwin = panel->inspected;
    WDDomain *db = panel->frame->screen_ptr->dwindow;
    proplist_t dict = db->dictionary;
    proplist_t def_win, window, value, key, Yes;
    char buffer[256], *icon_file;
    int modified = 0;
    int isDWin = 0;   /* true if we update for default window */

    if (WMGetButtonState(panel->instRb) != 0)
        key = PLMakeString(wwin->wm_instance);
    else if (WMGetButtonState(panel->clsRb) != 0)
        key = PLMakeString(wwin->wm_class);
    else if (WMGetButtonState(panel->bothRb) != 0) {
        strcat(strcat(strcpy(buffer, wwin->wm_instance), "."), wwin->wm_class);
        key = PLMakeString(buffer);
    }
#if 0
    else if (WMGetButtonState(panel->defaultRb) != 0) {
        key = PLMakeString("*");
        isDWin = 1;
    }
#endif
    else
        key = NULL;

    if (!dict) {
        dict = PLMakeDictionaryFromEntries(NULL, NULL, NULL);
        db->dictionary = dict;
        if (dict) {
            proplist_t tmp = PLMakeString(db->path);
            PLSetFilename(dict, tmp);
            PLRelease(tmp);
        }
    }

    if (!key || !dict)
        return;

    value = PLMakeString("*");
    def_win = PLGetDictionaryEntry(dict, value);
    PLRelease(value);

    window = PLMakeDictionaryFromEntries(NULL, NULL, NULL);
    Yes = PLMakeString("YES");

    /* Update icon for window */
    icon_file = WMGetTextFromTextField(panel->fileText);
    if (icon_file) {
        value = PLMakeString(icon_file);
        insertAttribute(def_win, window, "Icon", value, isDWin, &modified);
        PLRelease(value);
        free(icon_file);
    }
    if (WMGetButtonState(panel->alwChk))
        insertAttribute(def_win, window, "AlwaysUserIcon", Yes, isDWin, &modified);

    if (WMGetButtonState(panel->attrChk[0]))
        insertAttribute(def_win, window, "NoTitlebar", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[1]))
        insertAttribute(def_win, window, "NoResizebar", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[2]))
        insertAttribute(def_win, window, "NotClosable", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[3]))
        insertAttribute(def_win, window, "NotMiniaturizable", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[4]))
        insertAttribute(def_win, window, "KeepOnTop", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[5]))
        insertAttribute(def_win, window, "Omnipresent", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[6]))
        insertAttribute(def_win, window, "Unfocusable", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->attrChk[7]))
        insertAttribute(def_win, window, "SkipWindowList", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->moreChk[0]))
        insertAttribute(def_win, window, "NoHideOthers", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->moreChk[1]))
        insertAttribute(def_win, window, "NoKeyBindings", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->moreChk[2]))
        insertAttribute(def_win, window, "NoMouseBindings", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->moreChk[3]))
        insertAttribute(def_win, window, "KeepInsideScreen", Yes, isDWin, &modified);
    if (WMGetButtonState(panel->moreChk[4]))
        insertAttribute(def_win, window, "NoAppIcon", Yes, isDWin, &modified);

    if (WMGetButtonState(panel->curRb) != 0) {
        int n = panel->inspected->screen_ptr->current_workspace;
        WWorkspace *workspace = panel->inspected->screen_ptr->workspaces[n];
        if (workspace->name) {
            value = PLMakeString(workspace->name);
            insertAttribute(def_win, window, "StartWorkspace", Yes, isDWin, &modified);
            PLRelease(value);
        }
    }
    else if (WMGetButtonState(panel->setRb) != 0) {
        char *ws_name = WMGetTextFromTextField(panel->wsText);
        if (ws_name) {
            value = PLMakeString(ws_name);
            insertAttribute(def_win, window, "StartWorkspace", Yes, isDWin, &modified);
            PLRelease(value);
            free(ws_name);
        }
    }

    PLRemoveDictionaryEntry(dict, key);
    if (modified) {
        PLInsertDictionaryEntry(dict, key, window);
    }
    PLSave(dict, YES);

    PLRelease(key);
    PLRelease(window);
    PLRelease(Yes);

}


static void
applySettings(WMButton *button, InspectorPanel *panel)
{
    WWindow *wwin = panel->inspected;
    WWindowAttributes *wflags = &wwin->window_flags;
    int floating, skip_window_list;
    
    wflags->titlebar         = !WMGetButtonState(panel->attrChk[0]);
    wflags->resizable        = !WMGetButtonState(panel->attrChk[1]);
    wflags->closable         = !WMGetButtonState(panel->attrChk[2]);
    wflags->miniaturizable   = !WMGetButtonState(panel->attrChk[3]);
    floating                 =  WMGetButtonState(panel->attrChk[4]);
    wflags->omnipresent      =  WMGetButtonState(panel->attrChk[5]);
    wflags->focusable        = !WMGetButtonState(panel->attrChk[6]);
    skip_window_list         =  WMGetButtonState(panel->attrChk[7]);
    wflags->hide_others      = !WMGetButtonState(panel->moreChk[0]);
    wflags->bind_keys        = !WMGetButtonState(panel->moreChk[1]);
    wflags->bind_mouse       = !WMGetButtonState(panel->moreChk[2]);
    wflags->dont_move_off    =  WMGetButtonState(panel->moreChk[3]);
    wflags->application      = !WMGetButtonState(panel->moreChk[4]);
    wflags->always_user_icon =  WMGetButtonState(panel->alwChk);
    
    if (wflags->floating != floating) {
        int wlevel = ((wflags->floating = floating))
            ? NSFloatingWindowLevel : NSNormalWindowLevel;
        ChangeStackingLevel(wwin->frame->core, wlevel);
    }

    if (wflags->skip_window_list != skip_window_list) {
        int action = ((wflags->skip_window_list = skip_window_list))
            ? ACTION_REMOVE : ACTION_ADD;
        UpdateSwitchMenu(wwin->screen_ptr, wwin, action);
    }
        
    wWindowConfigureBorders(wwin);
#if 0
    {
        WApplication *wapp;

        wapp = wApplicationOf(wwin->main_window);
        if (wapp)
            wIconUpdate(wapp->app_icon->icon);
    }
#endif
}


static void
revertSettings(WMButton *button, InspectorPanel *panel)
{
    WWindow *wwin = panel->inspected;
    WMArgument chkArg = {WARG_SELECTED, (WMValue)0};
    int i, n, floating, skip_window_list;
    char *wm_instance = NULL;
    char *wm_class = NULL;

    if (WMGetButtonState(panel->instRb) != 0)
        wm_instance = wwin->wm_instance;
    else if (WMGetButtonState(panel->clsRb) != 0)
        wm_class = wwin->wm_class;
    else if (WMGetButtonState(panel->bothRb) != 0) {
        wm_instance = wwin->wm_instance;
        wm_class = wwin->wm_class;
    }
#if 0
    else if (WMGetButtonState(panel->defaultRb) != 0)
        wm_class = "*";
#endif

    wDefaultFillAttributes(wwin->screen_ptr, wm_instance, wm_class,
                           &wwin->window_flags);
    for (i=0; i < 8; i++) {
	switch (i) {
	 case 0:
            chkArg.value = (WMValue)!wwin->window_flags.titlebar;
	    break;
	 case 1:
	    chkArg.value = (WMValue)!wwin->window_flags.resizable;
	    break;
	 case 2:
	    chkArg.value = (WMValue)!wwin->window_flags.closable;
	    break;
	 case 3:
	    chkArg.value = (WMValue)!wwin->window_flags.miniaturizable;
	    break;
	 case 4:
            floating = WMGetButtonState(panel->attrChk[4]);
            if (wwin->window_flags.floating != floating) {
                int wlevel = (wwin->window_flags.floating != 0)
                    ? NSFloatingWindowLevel : NSNormalWindowLevel;
                ChangeStackingLevel(wwin->frame->core, wlevel);
            }
            chkArg.value = (WMValue)wwin->window_flags.floating;
	    break;
	 case 5:
	    chkArg.value = (WMValue)wwin->window_flags.omnipresent;
	    break;
	 case 6:
	    chkArg.value = (WMValue)!wwin->window_flags.focusable;
	    break;
	 case 7:
            skip_window_list = WMGetButtonState(panel->attrChk[7]);
            if (wwin->window_flags.skip_window_list != skip_window_list) {
                int action = (wwin->window_flags.skip_window_list != 0)
                    ? ACTION_REMOVE : ACTION_ADD;
                UpdateSwitchMenu(wwin->screen_ptr, wwin, action);
            }
	    chkArg.value = (WMValue)wwin->window_flags.skip_window_list;
	    break;
	}
        WMConfigureButton(panel->attrChk[i], &chkArg, 1);
    }
    for (i=0; i < 5; i++) {
	switch (i) {
	 case 0:
	    chkArg.value = (WMValue)!wwin->window_flags.hide_others;
	    break;
	 case 1:
	    chkArg.value = (WMValue)!wwin->window_flags.bind_keys;
	    break;
	 case 2:
	    chkArg.value = (WMValue)!wwin->window_flags.bind_mouse;
	    break;
	 case 3:
	    chkArg.value = (WMValue)wwin->window_flags.dont_move_off;
	    break;
	 case 4:
	    chkArg.value = (WMValue)!wwin->window_flags.application;
	    break;
	}
	WMConfigureButton(panel->moreChk[i], &chkArg, 1);
    }
    chkArg.value = (WMValue)wwin->window_flags.always_user_icon;
    WMConfigureButton(panel->alwChk, &chkArg, 1);

    wWindowConfigureBorders(wwin);
    showIconFor(W_SCREEN(button), panel, wm_instance, wm_class,
                UPDATE_TEXT_FIELD);

    workspaceArgs[0].value = (WMValue)NULL;
    n = wDefaultGetStartWorkspace(wwin->screen_ptr, wm_instance, wm_class);
    if (n >= 0 && n <= wwin->screen_ptr->workspace_count) {
        if (n == wwin->screen_ptr->current_workspace) {
            WMPerformButtonClick(panel->curRb);;
        }
        else {
            WMPerformButtonClick(panel->setRb);;
            workspaceArgs[0].value = (WMValue)wwin->screen_ptr->workspaces[n]->name;
        }
    }
    else {
        chkArg.value = (WMValue)0;
        WMConfigureButton(panel->curRb, &chkArg, 1);
        WMConfigureButton(panel->setRb, &chkArg, 1);
    }
    
    WMConfigureTextField(panel->wsText, workspaceArgs, COUNTARG(workspaceArgs));
}


static InspectorPanel*
createInspectorForWindow(WWindow *wwin)
{
    WScreen *scr = wwin->screen_ptr;

    InspectorPanel *panel;
    Window parent;
    char charbuf[128];
    int i, sel;

    panel = malloc(sizeof(InspectorPanel));
    if (!panel) {
	wWarning(_("not enough memory to open window inspector panel"));
	return NULL;
    }

    
    panel->inspected = wwin;

    panel->nextPtr = panelList;
    panelList = panel;
    

    sprintf(charbuf, "Inspecting  %s.%s", 
	    wwin->wm_instance ? wwin->wm_instance : "?",
	    wwin->wm_class ? wwin->wm_class : "?");

    parent = XCreateSimpleWindow(dpy, scr->root_win, 0, 0, PWIDTH, PHEIGHT, 
				 0, 0, 0);


    XSetTransientForHint(dpy, parent, wwin->client_win);

    panel->win = WMCreateWindow(scr->wmscreen, parent, panelArgs,
				COUNTARG(panelArgs));

    /**** create common stuff ****/
    
    /* command buttons */
    buttonArgs[0].value = (WMValue)30;
    buttonArgs[1].value = (WMValue)_("Save");
    panel->saveBtn = WMCreateCommandButton(panel->win, buttonArgs, 
					    COUNTARG(buttonArgs));
    WMBindEvent(panel->saveBtn, WEV_CLICK, saveSettings, panel);

    buttonArgs[0].value = (WMValue)100;
    buttonArgs[1].value = (WMValue)_("Apply");
    panel->applyBtn = WMCreateCommandButton(panel->win, buttonArgs, 
                                            COUNTARG(buttonArgs));
    WMBindEvent(panel->applyBtn, WEV_CLICK, applySettings, panel);


    buttonArgs[0].value = (WMValue)170;
    buttonArgs[1].value = (WMValue)_("Revert");
    panel->revertBtn = WMCreateCommandButton(panel->win, buttonArgs,
					     COUNTARG(buttonArgs));
    WMBindEvent(panel->revertBtn, WEV_CLICK, revertSettings, panel);
    
    /* page selection buttons */
    pageBtnArgs[0].value = (WMValue)70;
    pageBtnArgs[1].value = (WMValue)"1";
    panel->pageBtn1 = WMCreateButton(panel->win, WBTOnOff, pageBtnArgs,
					    COUNTARG(pageBtnArgs));
    WMBindEvent(panel->pageBtn1, WEV_CLICK, changePage, panel);
    
    pageBtnArgs[0].value = (WMValue)100;
    pageBtnArgs[1].value = (WMValue)"2";
    panel->pageBtn2 = WMCreateButton(panel->win, WBTOnOff, pageBtnArgs,
					    COUNTARG(pageBtnArgs));
    WMBindEvent(panel->pageBtn2, WEV_CLICK, changePage, panel);
    
    pageBtnArgs[0].value = (WMValue)130;
    pageBtnArgs[1].value = (WMValue)"3";
    panel->pageBtn3 = WMCreateButton(panel->win, WBTOnOff, pageBtnArgs,
					    COUNTARG(pageBtnArgs));
    WMBindEvent(panel->pageBtn3, WEV_CLICK, changePage, panel);
    
    pageBtnArgs[0].value = (WMValue)160;
    pageBtnArgs[1].value = (WMValue)"4";
    panel->pageBtn4 = WMCreateButton(panel->win, WBTOnOff, pageBtnArgs,
					    COUNTARG(pageBtnArgs));
    WMBindEvent(panel->pageBtn4, WEV_CLICK, changePage, panel);

    /**** window spec ****/
    specFrmArgs[0].value = (WMValue)_("Window Specification");
    panel->specFrm = WMCreateFrame(panel->win, specFrmArgs, 
				   COUNTARG(specFrmArgs));

    sel = 1;
    if (wwin->wm_class && wwin->wm_instance) {
	sprintf(charbuf, "%s.%s", wwin->wm_instance, wwin->wm_class);
	specRadArgs[0].value = (WMValue)18;
	specRadArgs[1].value = (WMValue)charbuf;
	specRadArgs[2].value = (WMValue)sel; sel=0;
	panel->bothRb = WMCreateRadioButton(panel->specFrm, specRadArgs, 
					    COUNTARG(specRadArgs));
    }

    if (wwin->wm_instance) {
	specRadArgs[0].value = (WMValue)38;
	specRadArgs[1].value = (WMValue)wwin->wm_instance;
	specRadArgs[2].value = (WMValue)sel; sel=0;
	panel->instRb = WMCreateRadioButton(panel->specFrm, specRadArgs, 
					    COUNTARG(specRadArgs));
    }

    if (wwin->wm_class) {
	specRadArgs[0].value = (WMValue)58;
	specRadArgs[1].value = (WMValue)wwin->wm_class;
	specRadArgs[2].value = (WMValue)sel; sel=0;
	panel->clsRb = WMCreateRadioButton(panel->specFrm, specRadArgs, 
					   COUNTARG(specRadArgs));
    }

#if 0
    specRadArgs[0].value = (WMValue)78;
    specRadArgs[1].value = (WMValue)_("Defaults for all windows");
    specRadArgs[2].value = (WMValue)sel; sel=0;
    panel->defaultRb = WMCreateRadioButton(panel->specFrm, specRadArgs,
                                           COUNTARG(specRadArgs));
#endif

    specLblArgs[0].value = (WMValue)_("The configuration will apply to all\n"
                                      "windows that have their WM_CLASS property"
                                      " set to the above selected\nname, when saved.");

    panel->specLbl = WMCreateLabel(panel->win, specLblArgs, 
				   COUNTARG(specLblArgs));
    
    /**** attributes ****/
    attrFrmArgs[0].value = (WMValue)_("Attributes");
    panel->attrFrm = WMCreateFrame(panel->win, attrFrmArgs,
				   COUNTARG(attrFrmArgs));

    for (i=0; i < 8; i++) {
	switch (i) {
	 case 0:
	    attrChkArgs[0].value = (WMValue)_("Disable titlebar");
            attrChkArgs[2].value = (WMValue)!wwin->window_flags.titlebar;
	    break;
	 case 1:
	    attrChkArgs[0].value = (WMValue)_("Disable resizebar");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.resizable;
	    break;
	 case 2:
	    attrChkArgs[0].value = (WMValue)_("Not closable");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.closable;
	    break;
	 case 3:
	    attrChkArgs[0].value = (WMValue)_("Not miniaturizable");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.miniaturizable;
	    break;
	 case 4:
	    attrChkArgs[0].value = (WMValue)_("Keep on top");
	    attrChkArgs[2].value = (WMValue)wwin->window_flags.floating;
	    break;
	 case 5:
	    attrChkArgs[0].value = (WMValue)_("Omnipresent");
	    attrChkArgs[2].value = (WMValue)wwin->window_flags.omnipresent;
	    break;
	 case 6:
	    attrChkArgs[0].value = (WMValue)_("Don't let it take focus");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.focusable;
	    break;
	 case 7:
	    attrChkArgs[0].value = (WMValue)_("Skip window list");
	    attrChkArgs[2].value = (WMValue)wwin->window_flags.skip_window_list;
	    break;
	}
	attrChkArgs[1].value = (WMValue)(20*(i+1));
	panel->attrChk[i] = WMCreateSwitchButton(panel->attrFrm, attrChkArgs,
                                                 COUNTARG(attrChkArgs));
    }
    

    
    /**** more attributes ****/
    attrFrmArgs[0].value = (WMValue)_("Advanced");
    panel->moreFrm = WMCreateFrame(panel->win, attrFrmArgs,
				   COUNTARG(attrFrmArgs));

    for (i=0; i < 5; i++) {
	switch (i) {
	 case 0:
	    attrChkArgs[0].value = (WMValue)_("Ignore HideOthers");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.hide_others;
	    break;
	 case 1:
	    attrChkArgs[0].value = (WMValue)_("Don't bind keyboard shortcuts");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.bind_keys;
	    break;
	 case 2:
	    attrChkArgs[0].value = (WMValue)_("Don't bind mouse clicks");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.bind_mouse;
	    break;
	 case 3:
	    attrChkArgs[0].value = (WMValue)_("Keep inside screen");
	    attrChkArgs[2].value = (WMValue)wwin->window_flags.dont_move_off;
	    break;
	 case 4:
	    attrChkArgs[0].value = (WMValue)_("No application icon");
	    attrChkArgs[2].value = (WMValue)!wwin->window_flags.application;
	    break;
	}
	attrChkArgs[1].value = (WMValue)(20*(i+1));
	panel->moreChk[i] = WMCreateSwitchButton(panel->moreFrm, attrChkArgs,
						 COUNTARG(attrChkArgs));
    }

    
    /* miniwindow/workspace */
    
    iconFrmArgs[0].value = (WMValue)_("Miniwindow Image");
    panel->iconFrm = WMCreateFrame(panel->win, iconFrmArgs,
				   COUNTARG(iconFrmArgs));
    panel->iconLbl = WMCreateLabel(panel->iconFrm, iconLblArgs,
                                   COUNTARG(iconLblArgs));
    
    updateIconBtnArgs[0].value = (WMValue)_("Default");
    updateIconBtnArgs[1].value = (WMValue)30;
    panel->defaultIconBtn = WMCreateCommandButton(panel->iconFrm,
                                                  updateIconBtnArgs,
                                                  COUNTARG(updateIconBtnArgs));
    WMBindEvent(panel->defaultIconBtn, WEV_CLICK, defaultIcon, panel);

    updateIconBtnArgs[0].value = (WMValue)_("Update");
    updateIconBtnArgs[1].value = (WMValue)65;
    panel->updateIconBtn = WMCreateCommandButton(panel->iconFrm, 
						 updateIconBtnArgs,
						 COUNTARG(updateIconBtnArgs));
    WMBindEvent(panel->updateIconBtn, WEV_CLICK, updateIcon, panel);    
    
    fileLblArgs[0].value = (WMValue)_("Icon file name:");
    panel->fileLbl = WMCreateLabel(panel->iconFrm, fileLblArgs,
				   COUNTARG(fileLblArgs));

    fileArgs[0].value = (WMValue)NULL;
    panel->fileText = WMCreateTextField(panel->iconFrm, fileArgs,
                                        COUNTARG(fileArgs));

    alwChkArgs[0].value = (WMValue)_("Ignore client supplied icon");
    alwChkArgs[1].value = (WMValue)wwin->window_flags.always_user_icon;
    panel->alwChk = WMCreateSwitchButton(panel->iconFrm, alwChkArgs,
					 COUNTARG(alwChkArgs));

    showIconFor(W_SCREEN(panel->alwChk), panel, wwin->wm_instance,
                wwin->wm_class, UPDATE_TEXT_FIELD);

    wsFrmArgs[0].value = (WMValue)_("Initial Workspace");
    panel->wsFrm = WMCreateFrame(panel->win, wsFrmArgs, COUNTARG(wsFrmArgs));

    workspaceArgs[0].value = (WMValue)NULL;
    i = wDefaultGetStartWorkspace(wwin->screen_ptr, wwin->wm_instance,
                                  wwin->wm_class);
    if (i >= 0 && i <= wwin->screen_ptr->workspace_count) {
        if (i == wwin->screen_ptr->current_workspace) {
            wsRad1Args[1].value = (WMValue)1;
            wsRad2Args[1].value = (WMValue)0;
        }
        else {
            wsRad1Args[1].value = (WMValue)0;
            wsRad2Args[1].value = (WMValue)1;
            workspaceArgs[0].value = (WMValue)wwin->screen_ptr->workspaces[i]->name;
        }
    }
    else {
        wsRad1Args[1].value = (WMValue)0;
        wsRad2Args[1].value = (WMValue)0;
    }

    wsRad1Args[0].value = (WMValue)_("Current workspace");
    panel->curRb = WMCreateRadioButton(panel->wsFrm, wsRad1Args,
				       COUNTARG(wsRad1Args));
    
    panel->setRb = WMCreateRadioButton(panel->wsFrm, wsRad2Args,
				       COUNTARG(wsRad2Args));

    panel->wsText = WMCreateTextField(panel->wsFrm, workspaceArgs,
				      COUNTARG(workspaceArgs));



    panel->frame = wManageInternalWindow(scr, parent, wwin->client_win, charbuf,
					 wwin->frame_x+wwin->frame->core->width/2,
					 wwin->frame_y+wwin->frame->top_width*2, 
					 PWIDTH, PHEIGHT);
    
    panel->frame->window_flags.closable = 1;
    wWindowUpdateButtonImages(panel->frame);
    wFrameWindowShowButton(panel->frame->frame, WFF_RIGHT_BUTTON);
    panel->frame->frame->on_click_right = destroyInspector;

    
    WMMapSubwidgets(panel->win);
    WMMapSubwidgets(panel->specFrm);
    WMMapSubwidgets(panel->attrFrm);
    WMMapSubwidgets(panel->moreFrm);
    WMMapSubwidgets(panel->iconFrm);
    WMMapSubwidgets(panel->wsFrm);

    WMPerformButtonClick(panel->pageBtn1);

    WMMapWidget(panel->win);
    
    return panel;
}


void
wShowInspectorForWindow(WWindow *wwin)
{
    InspectorPanel *panel;

    if (wwin->flags.inspector_open)
      return;

    wwin->flags.inspector_open = 1;
    panel = createInspectorForWindow(wwin);
    wwin->inspector = panel;
}
